'''' Useful funcs '''
import numpy as np
from sympy import diff, lambdify


def cart2pol(x, y):
    ''' Converts karthesian to cylindrial '''
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return (rho, phi)


def fit_uncertainty(pcov):
    return np.sqrt(np.diag(pcov))


def uncertainty(data, NB=0, return_result=False, print_result=True, graph=None):

    '''' Funkce na výpočet nejistot '''
    if len(data) <= 1:
        raise TypeError("Data must have MORE elements than 1!")

    stud_koef = [0, 6.3, 4.3, 3.18, 2.78, 2.57, 2.45, 2.37, 2.31, 2.26, 2.2, 2.14, 2.09, 2.04, 2.00, 1.98, 1.96]

    n = len(data)

    if n <= 10:
        k = stud_koef[n - 1]
    elif n <= 12:
        k = stud_koef[10]
    elif n <= 15:
        k = stud_koef[11]
    elif n <= 20:
        k = stud_koef[12]
    elif n <= 30:
        k = stud_koef[13]
    elif n <= 60:
        k = stud_koef[14]
    elif n <= 120:
        k = stud_koef[15]
    else:
        k = stud_koef[16]

    xs = 1 / n * sum(data)  # střední hodnota
    sumx = 0

    for x in data:
        sumx = sumx + (x - xs) ** 2  # suma (xi - xs)**2

    sigma = np.sqrt(1 / (n * (n - 1)) * sumx)  # výpočet sigmy
    NA = k * sigma  # Nejistota typu A
    N = np.sqrt(NA ** 2 + NB ** 2)  # Výpočet nejistoty

    if print_result == True:
        print(f"x = ({xs:.5f} $\pm$ {N:.5f})")

    if graph != None:
        fig, ax = graph

        ax.plot(range(1, len(data) + 1), data, ".", color="black", label="Data")
        ax.plot(range(1, len(data) + 1), [xs]*len(data), "-", color="red", label="Mean")
        ax.plot(range(len(data) + 2), [xs + N] * (len(data) + 2), "--", color="green", label="Stdev")
        ax.plot(range(len(data) + 2), [xs - N] * (len(data) + 2), "--", color="green")

    if return_result == True:
        return (xs, N)


def GetMinIndex(inputlist):
    '''get index of the minimum value in the list'''
    min_value = min(inputlist)

    # return the index of minimum value

    min_index = []

    for i in range(0, len(inputlist)):

        if min_value == inputlist[i]:
            min_index.append(i)

    return min_index


def max_diff(data):
    '''Function that returns biggest difference in dataset'''
    max_difference = float("-inf")  # Initialize max_difference to negative infinity

    for i, value in enumerate(data):
        diff = abs(value - i)  # Calculate the difference between value and index
        max_difference = max(max_difference, diff)  # Update max_difference

    return max_difference


def log_b(a, base):
    ''' logaritm from a of base b'''
    return np.log(a) / np.log(base)



def gauss(xdata, A=1, x0=0, sigma=0.1):
    '''Gauss function, returns ydata'''
    ydata = A * np.exp(-(xdata-x0)**2/(2*sigma**2))
    return ydata


def err_prop(func, variables, values, deviations):
    ''' error propagation deviation calculation function'''

    values_dict = {}
    for var, val in zip(variables, values):
        values_dict[str(var)] = val

    y_dev = 0
    for var, dev in zip(variables,deviations):
        derivative = diff(func, var)
        f = lambdify(variables, derivative)
        y_dev += (f(**values_dict)*dev)**2

    return np.sqrt(y_dev)




import tkinter as tk
from tkinter import ttk
''' Progress bar for loading with one input '''

class progressLoadBar(tk.Tk):
    ''' Progress bar for loading with one input '''
    def __init__(self):
        super().__init__()
        self.geometry("400x50")
        self.title("Loading...")
        self.progress = ttk.Progressbar(self, orient="horizontal", length=350, mode="determinate")
        self.progress.pack(pady=20)

    def load_update(self,new_value):
        ''' Change bar to specific value '''
        self.progress["value"] = new_value
        self.progress.start()
        self.progress.update()
        self.progress.stop()

class UpDownButton(tk.Frame):
    '''special type of button'''
    def __init__(self, master,value=5):
        super().__init__(master)

        self.font = value
        self.up_button = tk.Button(self, text="▲", width=1, height=1, command=self.increment)
        self.up_button.pack(side="top")

        self.label = tk.Label(self, width=1, height=1, text=self.font)
        self.label.pack(side="top")

        self.down_button = tk.Button(self, text="▼", width=1, height=1, command=self.decrement)
        self.down_button.pack(side="top")

    def increment(self):
        if self.font < 50:
            self.font += 1
            self.label.config(text=self.font)

    def decrement(self):
        if self.font > 1:
            self.font -= 1
            self.label.config(text=self.font)

    def get_val(self):
        return self.font

class EntryLock(tk.Frame):
    ''' Entry with checkbox that turns entry on/off, while off insertion of text is disabled. '''
    def __init__(self, master=None,text=None, background=None, width=4):
        super().__init__(master)
        self.locked = tk.BooleanVar()
        self.locked.set(False)
        self.config(bg=background)

        self.entry = tk.Entry(self, state=tk.DISABLED if self.locked.get()==False else tk.NORMAL, width=width)
        self.entry.grid(row=0, column=0)

        self.lock_checkbutton = tk.Checkbutton(self, text=text, variable=self.locked, command=self.update_state,bg=background)
        self.lock_checkbutton.grid(row=0, column=1)

    def update_state(self):
        self.entry.config(state=tk.DISABLED if self.locked.get() == False else tk.NORMAL)

class EntryWithPlaceholder(tk.Entry):
    ''' Entry that has grey text as background. '''
    def __init__(self, master=None, text="", background=None, width=None):
        super().__init__(master)
        self.placeholder = text
        self.placeholder_color = "grey"
        self.default_color = self["fg"]

        if background:
            self.configure(bg=background)

        if width:
            self.configure(width=width)

        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<FocusOut>", self.on_focus_out)

        self.add_placeholder()

    def add_placeholder(self):
        self.insert(0, self.placeholder)
        self.config(fg=self.placeholder_color)

    def on_focus_in(self, event):
        if self.get() == self.placeholder:
            self.delete(0, "end")
            self.config(fg=self.default_color)

    def on_focus_out(self, event):
        if not self.get():
            self.add_placeholder()

class InfoCircleLabel(tk.Label):
    ''' Label with question mark that show info when cursor hover above it '''
    def __init__(self, master, info_text):
        super().__init__(master)
        self.info_text = info_text
        self.info_window = None
        self.create_circle_icon()
        self.bind("<Enter>", self.show_info)
        self.bind("<Leave>", self.hide_info)

    def create_circle_icon(self):
        size = 24  # Adjust the size of the circle
        self.circle_icon = tk.Canvas(self, width=size, height=size, bg=None, highlightthickness=0)
        self.circle_icon.create_oval(4, 4, size-4, size-4, outline="gray", width=2)
        self.circle_icon.create_text(size // 2, size // 2, text="?", fill="gray", font=("Helvetica", 12, "bold"))
        self.circle_icon.pack()

    def show_info(self, event):
        if not self.info_window:
            self.info_window = tk.Toplevel(self.master)
            self.info_window.wm_overrideredirect(True)
            self.info_window.wm_geometry("+{}+{}".format(event.x_root, event.y_root))
            self.info_window_label = tk.Label(self.info_window, text=self.info_text, bg="lightgray")
            self.info_window_label.pack(ipadx=10, ipady=5)

    def hide_info(self, event):
        if self.info_window:
            self.info_window.destroy()
            self.info_window = None


import math
from matplotlib.patches import Rectangle
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors


def plot_colortable(colors, *, ncols=4, sort_colors=True):
    '''
    mcolors.BASE_COLORS
    mcolors.cnames
    mcolors.CSS4_COLORS
    mcolors.TABLEAU_COLORS
    '''

    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12

    # Sort colors by hue, saturation, value and name.
    if sort_colors is True:
        names = sorted(
            colors, key=lambda c: tuple(mcolors.rgb_to_hsv(mcolors.to_rgb(c))))
    else:
        names = list(colors)

    n = len(names)
    nrows = math.ceil(n / ncols)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + 2 * margin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin/width, margin/height,
                        (width-margin)/width, (height-margin)/height)
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows-0.5), -cell_height/2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(text_pos_x, y, name, fontsize=14,
                horizontalalignment='left',
                verticalalignment='center')

        ax.add_patch(
            Rectangle(xy=(swatch_start_x, y-9), width=swatch_width,
                      height=18, facecolor=colors[name], edgecolor='0.7')
        )

    return fig


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.2.4'
DESCRIPTION = 'Useful functions and constants for Fyzing struggle'
LONG_DESCRIPTION = 'A package that helped me at University - Radim Slovák, VUT.'

# Setting up
setup(
    name="RSFlib",
    version=VERSION,
    author="Radzym (Radim Slovák)",
    author_email="<slovak.radim@seznam.cz>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    package_data={'RSFlib': ['opt_data/*.rsfile']},
    install_requires=["pandas"],
    keywords=['python', 'data analysis', "physics", "ellipsometry"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import pkg_resources
import numpy as np
import pandas as pd
from numpy import arctan, sqrt, pi, exp, log, array
from scipy.special import dawsn
from RSFlib.Constants import PhysConst

def get_MSE(e_psi, e_delta, f_psi, f_delta, M=3):
    '''
    Calculates mean-square-error
    :param e_psi: experimental psi
    :param e_delta: esperimental delta
    :param f_psi: model generated psi
    :param f_delta: model generated delta
    :param M: number of fiting parameterrs
    :return: MSE as reliability of fit
    '''

    N = len(e_psi)
    mse = 0
    sigma_psi = np.std(e_psi)
    sigma_delta = np.std(e_delta)
    for n in range(N):
        mse += ( ( (f_psi[n]-e_psi[n])/sigma_psi )**2 + ( (f_delta[n]-e_delta[n])/sigma_delta )**2 )
    return sqrt(1/(2*N-M) * mse)

def cauchy(wvl, A=0, B=0, C=0):
    '''
    Cauchy function for fitting real part of dielectric function
    :param wvl: wavelength in nm
    :param A: amplitude
    :param B:
    :param C:
    :return: real part of dielectric function
    '''

    wvl = wvl / 1000
    return A + B / wvl ** 2 + C / wvl ** 4

def tauc_lorentz(E, An, En, Cn, Eg, eps1_inf=0):
    '''
    tauc-lorentz function for fiting complex dielectric function
    :param E: energy in Ev
    :param An: amplitude in Ev
    :param En:
    :param Cn:
    :param Eg: band gab energy in Ev
    :param eps1_inf: offset in Ev
    :return:
    '''

    eps2 = (An*En*Cn*(E-Eg)**2) / ((E**2 - En**2)**2+Cn**2*E**2) * 1/E
    eps2 = np.array( [0 if E <= Eg else eps for eps, E in zip(eps2, E)] )

    # E_0 = En
    alp = sqrt(4*En**2 - Cn**2)
    gam = sqrt(En**2 - Cn**2/2)

    a_ln = (Eg**2-En**2)*E**2 + Eg**2*Cn**2 - En**2*(En**2 + 3*Eg**2)
    a_atan = (E**2 - En**2)*(En**2 + Eg**2) + Eg**2*Cn**2
    zeta4 = (E**2 - gam**2)**2 + alp**2*Cn**2/4

    eps1 = (
        eps1_inf + An*Cn/pi/zeta4 * a_ln/2/alp/En * log( (En**2 + Eg**2 + alp*Eg) / (En**2 + Eg**2 - alp*Eg) )
        - An/pi/zeta4 * a_atan/En * ( pi - arctan( (2*Eg + alp)/Cn ) + arctan( (-2*Eg + alp)/Cn ) )
        + 2*An*En/pi/zeta4/alp * Eg * (E**2 - gam**2) * ( pi + 2*arctan( 2 * (gam**2-Eg**2)/alp/Cn ) )
        - An*En*Cn/pi/zeta4 * (E**2 + Eg**2)/E * log( abs(E-Eg) / (E+Eg) )
        + 2*An*En*Cn/pi/zeta4 * Eg * log( (abs(E-Eg)*(E+Eg)) / sqrt( (En**2 - Eg**2)**2 + Eg**2*Cn**2 ) )
    )

    return np.vectorize(complex)(eps1, eps2)

def gauss(E, An, Br, En, eps_inf=0):
    '''
    gauss function for complex dielectric function
    :param E: energy in Ev
    :param An: Amplitude in Ev
    :param Br: half of Width till konvergention to 0 in Ev
    :param En:
    :param eps_inf:
    :return:
    '''

    E = np.array(E)
    sigma = Br/2/sqrt(log(2))
    eps2 = An * exp(-( (E-En)/sigma )**2) - An * exp(-( (E+En)/sigma )**2)

    eps1 = eps_inf + 2*An/sqrt(pi) * ( dawsn(2 * sqrt(log(2)) * (E + En)/Br) - dawsn(2 * sqrt(log(2)) * (E - En)/Br) )

    return np.vectorize(complex)(eps1, eps2)

def selmaier(E, A_1, E_1):
    return A_1 / (E_1 ** 2 - E ** 2)

def calculate_perm(n, k):
    ''' function that makes permitivity from refractive index and absor '''

    epsilon1 = n**2 - k**2
    epsilon2 = 2 * n * k
    return epsilon1, epsilon2

def calculate_nk(epsilon1, epsilon2):
    ''' function that makes nk from permitivity '''

    n = np.sqrt((epsilon1 + np.sqrt(epsilon1**2 + epsilon2**2)) / 2)
    k = np.sqrt((np.sqrt(epsilon1**2 + epsilon2**2) - epsilon1) / 2)
    return n, k

def Ev2m(data):
    ''' calculates Ev to m '''
    pc = PhysConst()
    h, e, c = pc.planck, pc.elem_charge, pc.light_speed
    return h*c/data/e

def elips_eps(psi, delta, theta):
    ''' calculates complex dielectric function from psi, delta and theta, must be in radians '''
    psi = array(psi)
    delta = array(delta)
    theta = array(theta)
    rho = np.tan(psi) * np.exp(1j * delta)
    eps = np.sin(theta) ** 2 * (1 + ((1 - rho) / (1 + rho)) ** 2 * np.tan(theta) ** 2)
    return eps

def get_opt_const(input_energies, data_file, from_library=False, delimiter="\t"):
    ''' get you data acording to Ev, the had has to have eV, n, k

    options:
        al2o3, Sb3O2_FL_fit, si_jaw2, sio2_jaw2, sio2_fmetrix
    '''

    if from_library == False:
        df = pd.read_csv(f"opt_data/{data_file}.rsfile", delimiter=delimiter)

    else:
        file_path = pkg_resources.resource_filename('RSFlib.opt_data', f'{data_file}.rsfile')
        df = pd.read_csv(file_path, delimiter=delimiter)


    # Function to find the closest energy in the data to a given input energy
    def find_closest_energy(input_energy):
        closest_row = df.iloc[(df['eV'] - input_energy).abs().idxmin()]
        return complex(closest_row['n'], closest_row["k"])

    refractive_indexes = np.array([find_closest_energy(energy) for energy in input_energies])
    return refractive_indexes